#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>
#include "CommandQueue.h"
#include "log.h"
#include "stdafx.h"


//
// simple utility text processing functions:
//
//   newstrcpy( s ) -- allocate memory for and create a copy of a string
//   skipwhite( s ) -- return pointer to next non-white character
//   skipnonwhite( s ) -- return pointer to next white character
//   zapchar( s, c ) -- look for 'c', replace it with 0, return pointer to next character
//
char* newstrcpy( char* s ) {
  char* result = (char*)malloc( strlen( s ) + 1 );
  strcpy( result, s );
  return( result );
}

char* skipwhite( char* scanner ) {
  while ( *scanner ) {
    if ( isspace( *scanner )) {
      scanner++;
    } else {
      return( scanner );
    }
  }
  return( scanner );
}

char* skipnonwhite( char* scanner ) {
	bool inquote = false;
	while ( *scanner ) {
		if ( *scanner == '"' ) {
			if ( inquote ) {
				inquote = false;
			} else {
				inquote = true;
			}
		}
		if ( isspace( *scanner )) {
			if ( !inquote ) {
				return( scanner );
			}
		}
		scanner++;
	}
	return( scanner );
}

char* zapchar( char* scanner, char z ) {
  while ( *scanner ) {
    if ( *scanner == z ) {
      *scanner = 0;
      return( scanner + 1 );
    } else {
      scanner++;
    }
  }
  return( NULL );
}


//
//  general classes:
//
//     dwTokenizer -- name=value tokens, where "=value" is optional
//     dwLanguageLine -- type name=value ..., where "name=value" is repeated several times
//     dwGrid -- handles grid specific language syntax with default values
//     dwImage -- handles image specific language syntax with default values
//

#define dwMAX 10

class dwTokenizer {
  protected:
    char* token[ dwMAX ];
    char* name[ dwMAX ];
    char* value[ dwMAX ];
    int numberTokens;
  public:
    dwTokenizer( char* line );
    char* getToken( int i ) { return( token[ i ] ); }
    char* getName( int i ) { return( name[ i ] ); }
    char* getValue( int i ) { return( value[ i ] ); }
    int getNumberTokens() { return( numberTokens ); }
};

dwTokenizer::dwTokenizer( char* line ) {
  for ( int i = 0; i < dwMAX; i++ ) {
    token[ i ] = name[ i ] = value[ i ] = NULL;
  }
  numberTokens = 0;
  char* scanner = skipwhite( line );
  while ( *scanner ) {
    token[ numberTokens++ ] = scanner;
    scanner = skipnonwhite( scanner );
    if ( *scanner ) {
      *scanner = 0;
      scanner = skipwhite( scanner + 1 );
    }
  }
  for ( i = 0; i < numberTokens; i++ ) {
    name[ i ] = token[ i ];
    value[ i ] = zapchar( name[ i ], '=' );
  }
}
 

class dwLanguageLine : dwTokenizer {
  private:
    char* dwType;
  public:
    dwLanguageLine( char* line );
    char* getType() { return( dwType ); }
    char* getName( int i ) { return( name[ i ] ); }
    char* getValue( int i ) { return( value[ i ] ); }
    char* getValue( char* name );
};

dwLanguageLine::dwLanguageLine( char* line ) : dwTokenizer( line ) {
  dwType = getToken( 0 );
}

char* dwLanguageLine::getValue( char* name ) {
  int nTokens = getNumberTokens();
  for ( int i = 0; i < nTokens; i++ ) {
    if ( strcmp( name, getName( i )) == 0 ) {
      return( getValue( i ));
    }
  }
  return( NULL );
}
  
class dwGrid : public dwLanguageLine {
  private:
	// size=AxB where A and B are int values > 0
    int width;
    int height;

	// color=r,g,b (where r,g,b is 0-1 range of vlaues)
    char* red;
    char* green;
    char* blue;

	// location=x,y,z
	char* x;
	char* y;
	char* z;

	// name=val
	char* name;
  public:
    dwGrid( char* line );
    int getWidth() { return( width ); }
    int getHeight() { return( height ); }
    char* getRed() { return( red ); }
    char* getGreen() { return( green ); }
    char* getBlue() { return( blue ); }
	char* getName() { return( name ); }
    void generate( CommandQueue* cq );
};

class dwLabel : public dwLanguageLine {
private:
	// location=x,y,z
	char* x;
	char* y;
	char* z;

	// color=r,g,b
	char* red;
	char* green;
	char* blue;

	// text="the test label goes here"
	char* text;

	// name=val
	char* name;

	// scale=n
	char* scale;
public:
	dwLabel( char* line );
	char* getX() { return( x ); }
	char* getY() { return( y ); }
	char* getZ() { return( z ); }
	char* getRed() { return( red ); }
	char* getGreen() { return( green ); }
	char* getBlue() { return( blue ); }
	char* getName() { return( name ); }
	char* getText() { return( text ); }
	void generate( CommandQueue* cq );
};

class dwView : public dwLanguageLine {
  private:
	// location=x,y,z
    char* x;
	char* y;
	char* z;

	// name=val
	char* name;
  public:
    dwView( char* line );
    char* getX() { return( x ); }
    char* getY() { return( y ); }
    char* getZ() { return( z ); }
	char* getName() { return( name ); }
    void generate( CommandQueue* cq );
};

class dwImage: public dwLanguageLine {
  private:
	// location=x,y,z
    char* x;
	char* y;
	char* z;

	// image URL
	char* url;

	// size=AxB where A and B are int values > 0
    char* width;
    char* height;

	// name=val
	char* name;
  public:
    dwImage( char* line );
    char* getX() { return( x ); }
    char* getY() { return( y ); }
    char* getZ() { return( z ); }
	char* getName() { return( name ); }
    void generate( CommandQueue* cq );
};

dwImage::dwImage( char* line ) : dwLanguageLine( line ) {
  x = "0";
  y = "0";
  z = "0";
  char* location = getValue( "location" );
  if ( location != NULL ) {
    char* xs = location;
    char* ys = zapchar( xs, ',' );
	char* zs = zapchar( ys, ',' );
	if ( xs != NULL ) {
		x = xs;
	}
	if ( ys != NULL ) {
		y = ys;
	}
	if ( zs != NULL ) {
		z = zs;
	}
  }
  name = "image";
  char* lname = getValue( "name" );
  if ( lname != NULL ) {
	  name = lname;
  }
  url = "c:/mindtel/render/textures/aworld.png";
  char* iurl = getValue( "url" );
  if ( iurl != NULL ) {
	  url = iurl;
  }
  width = "1";
  height = "1";
  char* size = getValue( "size" );
  if ( size != NULL ) {
    char* w = size;
    char* h = zapchar( w, 'x' );
	if ( w != NULL ) {
		width = w;
	}
	if ( h != NULL ) {
		height = h;
	}
  }
}

void dwImage::generate( CommandQueue* cq) {
  cq->addCommand( Data, "Transform { translation " );
  char lbuf[30];
  sprintf( lbuf, "%s %s %s ", x, y, z );
  cq->addCommand( Data, lbuf );
  cq->addCommand( Data, " children [ Shape { appearance Appearance { " );
  char ibuf[200];
  sprintf( ibuf, " texture ImageTexture { url \"%s\" } } ", url );
  cq->addCommand( Data, ibuf );
  char coordlocs[100];
  sprintf( coordlocs, "0 0 0, 0 %s 0, %s %s 0, %s 0 0", width, width, height, height );
  cq->addCommand( Data, " geometry IndexedFaceSet { coord Coordinate { point [ " );
  cq->addCommand( Data, coordlocs );
  cq->addCommand( Data, " ] }" );
  cq->addCommand( Data, " coordIndex [ 3 2 1 0 -1 ] } " );
  cq->addCommand( Data, " } ] }" );
}


dwLabel::dwLabel( char* line ) : dwLanguageLine( line ) {
  x = "0";
  y = "0";
  z = "0";
  char* location = getValue( "location" );
  if ( location != NULL ) {
    char* xs = location;
    char* ys = zapchar( xs, ',' );
	char* zs = zapchar( ys, ',' );
	if ( xs != NULL ) {
		x = xs;
	}
	if ( ys != NULL ) {
		y = ys;
	}
	if ( zs != NULL ) {
		z = zs;
	}
  }
  scale = "1";
  char* tscale = getValue( "scale" );
  if ( tscale != NULL ) {
	  scale = tscale;
  }

  char* color = getValue( "color" );
  red = "1";
  green = "1";
  blue = "1";
  if ( color != NULL ) {
	  red = color;
	  green = zapchar( red, ',' );
	  blue = zapchar( green, ',' );
  }
  name = "label";
  char* lname = getValue( "name" );
  if ( lname != NULL ) {
	  name = lname;
  }
  text = "test label";
  char* label = getValue( "text" );
  if ( label != NULL ) {
	  text = label;
  }
}

void dwLabel::generate( CommandQueue* cq) {
  cq->addCommand( Data, "Transform { translation " );
  char lbuf[30];
  sprintf( lbuf, "%s %s %s ", x, y, z );
  cq->addCommand( Data, lbuf );
  cq->addCommand( Data, " scale " );
  sprintf( lbuf, "%s %s %s ", scale, scale, scale );
  cq->addCommand( Data, lbuf );
  cq->addCommand( Data, " children [ Shape { appearance Appearance { " );
  char ibuf[200];
  sprintf( ibuf, " material Material { emissiveColor %s %s %s  diffuseColor %s %s %s } } ", red, green, blue, red, green, blue );
  cq->addCommand( Data, ibuf );
  sprintf( ibuf, " geometry DEF %s_text Text3D { string [ ", name );
  cq->addCommand( Data, ibuf );
  cq->addCommand( Data, text );
  cq->addCommand( Data, " ] }" );
  cq->addCommand( Data, " } " );
  sprintf( ibuf, " DEF Script_%s Script {", name );
  cq->addCommand( Data, ibuf );
  cq->addCommand( Data, " eventIn SFString ins1" );
  cq->addCommand( Data, " eventOut MFString outString" );
  cq->addCommand( Data, " url \"javascript:" );
  cq->addCommand( Data, "   function ins1(value) {" );
  cq->addCommand( Data, "     outString[0] = value;" );
  cq->addCommand( Data, "  }\" }" );
  sprintf( ibuf, " ROUTE Script_%s.outString TO %s_text.set_string", name, name );
  cq->addCommand( Data, ibuf );
  cq->addCommand( Data, " ] }" );
}

dwView::dwView( char* line ) : dwLanguageLine( line ) {
  x = "0";
  y = "0";
  z = "10";
  char* location = getValue( "location" );
  if ( location != NULL ) {
    char* xs = location;
    char* ys = zapchar( xs, ',' );
	char* zs = zapchar( ys, ',' );
	if ( xs != NULL ) {
		x = xs;
	}
	if ( ys != NULL ) {
		y = ys;
	}
	if ( zs != NULL ) {
		z = zs;
	}
  }
  name = "view";
  char* lname = getValue( "name" );
  if ( lname != NULL ) {
	  name = lname;
  }
}


void dwView::generate( CommandQueue* cq) {
  cq->addCommand( Data, "Viewpoint { " );
  char buf[100];
  sprintf( buf, "position %s %s %s", x, y, z );
  cq->addCommand( Data, buf );
  cq->addCommand( Data, "} ");
}

dwGrid::dwGrid( char* line ) : dwLanguageLine( line ) {
  x = "0";
  y = "0";
  z = "0";
  char* location = getValue( "location" );
  if ( location != NULL ) {
    char* xs = location;
    char* ys = zapchar( xs, ',' );
	char* zs = zapchar( ys, ',' );
	if ( xs != NULL ) {
		x = xs;
	}
	if ( ys != NULL ) {
		y = ys;
	}
	if ( zs != NULL ) {
		z = zs;
	}
  }

  width = 10;
  height = 10;
  char* size = getValue( "size" );
  if ( size != NULL ) {
    char* w = size;
    char* h = zapchar( w, 'x' );
    int wint = atoi( w );
    int hint = atoi( h );
    if ( wint > 0 ) {
      width = wint;
    }
    if ( hint > 0 ) {
      height = hint;
    }
  }
  red = ".5";
  green = ".5";
  blue = ".5";
  char* rgb = getValue( "color" );
  if ( rgb != NULL ) {
    char* r = rgb;
    char* g = zapchar( r, ',' );
    char* b = zapchar( g, ',' );
    if ( r != NULL ) {
      red = r;
    }
    if ( g != NULL ) {
      green = g;
    }
    if ( b != NULL ) {
      blue = b;
    }
  }
  name = "grid";
  char* lname = getValue( "name" );
  if ( lname != NULL ) {
	  name = lname;
  }

}

void dwGrid::generate( CommandQueue* cq) {
  int i, j;
  cq->addCommand( Data, "Transform { " );
  char cbuf[30];
  sprintf( cbuf, "translation %s %s %s ", x, y, z );
  cq->addCommand( Data, cbuf );
  cq->addCommand( Data, " children [ Shape { appearance Appearance { material Material { " );
  sprintf( cbuf, " emissiveColor %s %s %s } }", red, green, blue );
  cq->addCommand( Data, cbuf );
  cq->addCommand( Data, " geometry IndexedLineSet { coord Coordinate { point [\n" );
  for ( i = 0; i < width + 1; i++ ) {
    for ( j = 0; j < height + 1; j++ ) {
      char buf[50];
      sprintf( buf, "%d %d 0, ", i, j );
      cq->addCommand( Data, buf );
    }
  }
  cq->addCommand( Data, " ] } coordIndex [\n" );
  for ( j = 0; j < width + 1; j++ ) {
    for ( i = 0; i < height; i++ ) {
      char buf[50];
      sprintf( buf, "%d %d -1, ", (height+1)*j + i, (height+1)*j + i + 1 );
      cq->addCommand( Data, buf );
    }
  }
  for ( i = 0; i < width; i++ ) {
    for ( j = 0; j < height + 1; j++ ) {
      char buf[50];
      sprintf( buf, "%d %d -1, ", i*(height+1) + j, (i+1)*(height+1) + j );
	  cq->addCommand( Data, buf );
    }
  }
  cq->addCommand( Data, " ] } } ] }" );
}

void dwHandle( CommandQueue* cq, char* line ) {
	if ( strncmp( line, "grid", 4 ) == 0 ) {
		dwGrid dwg( line );
		cq->suspend();
		cq->addCommand( StartData, dwg.getName() );
		dwg.generate( cq );
		cq->addCommand( EndData, NULL );
		cq->resume();
	} else if ( strncmp( line, "view", 4 ) == 0 ) {
		dwView dwv( line );
		cq->suspend();
		cq->addCommand( StartData, dwv.getName() );
		dwv.generate( cq );
		cq->addCommand( EndData, NULL );
		cq->resume();
	} else if ( strncmp( line, "image", 5 ) == 0 ) {
		dwImage dwi( line );
		cq->suspend();
		cq->addCommand( StartData, dwi.getName() );
		dwi.generate( cq );
		cq->addCommand( EndData, NULL );
		cq->resume();
	} else if ( strncmp( line, "label", 5 ) == 0 ) {
		dwLabel dwl( line );
		cq->suspend();
		cq->addCommand( StartData, dwl.getName() );
		dwl.generate( cq );
		cq->addCommand( EndData, NULL );
		cq->resume();
	} else if ( strncmp( line, "link", 4 ) == 0 ) {
		cq->suspend();
		cq->addCommand( Link, line + 5 );
		cq->resume();
	}
}